import re
from typing import Callable, List, Optional

import MeCab

from jadoc.errors import InvalidTokenizerError, NotFoundNodeFormatError
from jadoc.word.word import Word


def _mecab_tagger(
    dicdir: Optional[str] = None,
    node_format: Optional[str] = None,
    unk_format: Optional[str] = None,
) -> MeCab.Tagger:
    """Generate MeCab.Tagger

    Parameters
    ----------
    dicdir : str, optional
        Path of MeCab dictionary directory (the default is to not specify).
    node_format : str, optional
        MeCab ``node_format`` (the default is to not specify).
    unk_format : str, optional
        MeCab ``unk_format`` (the default is to not specify).

    Returns
    -------
    MeCab.Tagger
        MeCab.Tagger with optional arguments.

    See Also
    --------
    ``mecab --help``
    """
    options = []

    if dicdir is not None:
        options.append("--dicdir=" + str(dicdir))

    if node_format is not None:
        options.append("--output-format-type=")  # Dealing with the MeCab bug.
        options.append("--node-format=" + node_format)

    if unk_format is not None:
        options.append("--unk-format=" + unk_format)

    return MeCab.Tagger(" ".join(options))


def _find_index(
    items: List[str], equal_to: Optional[str] = None, include: Optional[str] = None
) -> Optional[int]:
    """Find the index of the first item in an array that satisfies a particular
    condition.

    Parameters
    ----------
    items : list of str
        Search target.
    equal_to : str, optional
        If specified, this function will search the item matching ``equal_to``.
    include : str, optional
        If specified, this function will search the item containing ``include``.

    Returns
    -------
    int or None
        index, or None if not found.
    """
    for i, item in enumerate(items):
        if equal_to is not None and item == equal_to:
            return i
        if include is not None and include in item:
            return i
    return None


def _find_node_format(dicdir: Optional[str] = None) -> Optional[str]:
    """Find the MeCab ``node_format`` automatically.

    Parameters
    ----------
    dicdir : str, optional
        Path of MeCab dictionary directory (the default is to not specify).

    Returns
    -------
    str or None
        MeCab ``node_format``, or None if not found.

    Raises
    ------
    NotFoundNodeFormatError
        If the ``node-format`` could not be set automatically.
    """
    tagger = _mecab_tagger(dicdir=dicdir, node_format=r"%H\\n")

    text = "こおりつけ"
    csv = tagger.parse(text).splitlines()[0]
    items = re.split(r"[\t,]", csv)

    pos_i = _find_index(items, equal_to="動詞")
    base_i = _find_index(items, equal_to="こおりつく")
    c_type_i = _find_index(items, include="五段")
    c_form_i = _find_index(items, include="命令")

    if any([i is None for i in (pos_i, base_i, c_type_i, c_form_i)]):
        msg = (
            "Please configure ``node-format`` manually to match the format "
            + r"``node-format=surface,pos,baseForm,cType,cForm\\n``"
        )
        raise NotFoundNodeFormatError(msg)

    node_format = (
        ",".join(
            [
                r"%m",  # surface
                f"%F-[{pos_i},{pos_i + 1},{pos_i + 2},{pos_i + 3}]",  # pos
                f"%f[{base_i}]",  # baseForm
                f"%f[{c_type_i}]",  # cType
                f"%f[{c_form_i}]",  # cForm
            ]
        )
        + r"\\n"
    )

    return node_format


def check_tokenizer(tokenize: Callable[[str], List[Word]]) -> None:
    """Check that the tokenize function is working properly.

    Parameters
    ----------
    tokenize : function
        Tokenize function.

    Returns
    -------
    None
        Return None when the tokenize function is working properly.

    Raises
    ------
    InvalidTokenizerError
        If the tokenize function is not working properly.
    """
    tokens = ["楽しい", "本", "を", "よく", "読み", "ます"]
    conj_tokens = ["楽しい", "読み", "ます"]
    words = tokenize("".join(tokens))
    if len(words) != len(tokens):
        raise InvalidTokenizerError("/".join([word.surface for word in words]))
    for word, token in zip(words, tokens):
        if word.surface != token:
            raise InvalidTokenizerError(str(word))
        if "詞" not in word.pos.value[0]:
            raise InvalidTokenizerError(str(word.pos))
        if word.has_conjugation and word.surface not in conj_tokens:
            raise InvalidTokenizerError(str(word))


def generate_tokenizer(
    dicdir: Optional[str] = None, node_format: Optional[str] = None
) -> Callable[[str], List[Word]]:
    """Generate a function that converts the text into a list of Word objects.

    Parameters
    ----------
    dicdir : str, optional
        Path of MeCab dictionary directory (the default is delegated to MeCab).
    node_format : str, optional
        MeCab ``node_format`` that matches the format
        ``node-format=surface,pos,baseForm,cType,cForm\\n``.
        If not specified, the system will try to find it automatically.

    Returns
    -------
    function
        A function that converts the text into a list of Word objects.

    See Also
    --------
    jadoc.word.word.Word :
        A class that represents each word.

    Examples
    --------
    >>> tokenize = generate_tokenizer()
    >>> words = tokenize("毎日とても歩きます")
    >>> assert [word.surface for word in words] == ["毎日", "とても", "歩き", "ます"]
    """
    if node_format is None:
        node_format = _find_node_format(dicdir)

    mecab_tagger = _mecab_tagger(dicdir=dicdir, node_format=node_format)

    blank = ("", " ", "　")

    def _tokenize(text: str) -> List[Word]:
        parsed = mecab_tagger.parse(text)

        if parsed is None:
            return [Word(surface=text, pos_info=["名詞"])]

        nodes = parsed.splitlines()[:-1]

        words = []
        for node in nodes:
            attrs = node.split(",")

            surface = attrs[0]
            pos_info = attrs[1].split("-")
            base = attrs[2]
            c_type_info = attrs[3]
            c_form_info = attrs[4]

            if c_type_info in blank and c_form_info in blank:
                word = Word(surface=surface, pos_info=pos_info, base=base)
            else:
                word = Word(
                    surface=surface,
                    pos_info=pos_info,
                    base=base,
                    c_type_info=c_type_info,
                    c_form_info=c_form_info,
                )

            words.append(word)

        return words

    check_tokenizer(_tokenize)

    return _tokenize



import os

from . import __title__

ENV_DEBUG = f"{__title__.upper()}_DEBUG"


def debug_on() -> bool:
    """Whether debug mode is enabled or not.

    Returns
    -------
    bool
        True if debug mode is enabled, False otherwise.
    """
    env = os.getenv(ENV_DEBUG)
    if env is None:
        # default
        return False
    if env.lower().strip() in ("true", "on"):
        return True
    else:
        return False



from typing import Dict, List, Optional

from .cform import get_normalized_cform
from .ctype import Nothing, Unknown, get_normalized_ctype
from .pos import get_normalized_pos


class Word:
    def __init__(
        self,
        surface: str,
        pos_info: List[str],
        base: Optional[str] = None,
        c_type_info: Optional[str] = None,
        c_form_info: Optional[str] = None,
    ) -> None:
        self.surface = surface
        self.pos = get_normalized_pos(pos_info)

        if base is None or base == "":
            self.base = surface
        else:
            self.base = base

        if c_type_info is None:
            c_type_info = ""
        self.c_type = get_normalized_ctype(pos_info, self.base, c_type_info)

        if c_form_info is None:
            c_form_info = ""
        self.c_form = get_normalized_cform(surface, pos_info, c_form_info)

        self.has_conjugation = (
            type(self.c_type) != Nothing and type(self.c_type) != Unknown
        )

    def to_dict(self) -> Dict[str, str]:
        """Convert this word to a ``dict`` object.

        Returns
        -------
        dict
            The attributes of this word.
        """
        return {
            "surface": self.surface,
            "pos": str(self.pos),
            "base": self.base,
            "c_type": str(self.c_type),
            "c_form": str(self.c_form),
            "has_conjugation": str(self.has_conjugation),
        }

    def __str__(self) -> str:
        """Represent this word as a string.

        Returns
        -------
        str
            A string representing the attributes of this word.
        """
        return str(self.to_dict())



from abc import ABCMeta, abstractmethod
from typing import List, Type

from jadoc.errors import CannotConjugateError

from . import pos


class ConjugationType(metaclass=ABCMeta):
    """
    See Also
    --------
    ``https://www.sketchengine.eu/tagset-jp-mecab/``
    """

    def __init__(self, value: str) -> None:
        self.value = value

    @staticmethod
    @abstractmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        pass

    @staticmethod
    @abstractmethod
    def conjugate(base: str, ending: str) -> str:
        """
        活用した結果の表層形を得る。
        """
        pass

    def __str__(self) -> str:
        return self.__class__.__name__ + "(" + self.value + ")"


class Godan(ConjugationType):
    """
    五段活用
    """

    POLITE = ("ござる", "御座る", "なさる", "為さる", "くださる", "下さる", "おっしゃる", "仰る", "いらっしゃる")
    JA_CHARS = {
        "a": ("か", "さ", "た", "な", "は", "ま", "ら", "わ", "が", "ざ", "だ", "ば"),
        "i": ("き", "し", "ち", "に", "ひ", "み", "り", "い", "ぎ", "じ", "ぢ", "び"),
        "u": ("く", "す", "つ", "ぬ", "ふ", "む", "る", "う", "ぐ", "ず", "づ", "ぶ"),
        "e": ("け", "せ", "て", "ね", "へ", "め", "れ", "え", "げ", "ぜ", "で", "べ"),
        "o": ("こ", "そ", "と", "の", "ほ", "も", "ろ", "お", "ご", "ぞ", "ど", "ぼ"),
    }

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return base != "ある" and "五段" in c_type_info

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        e = ending[0]
        if e not in Godan.JA_CHARS.keys():
            return base[:-1] + e
        u = base[-1]
        idx = Godan.JA_CHARS["u"].index(u)
        ending = Godan.JA_CHARS[e][idx] + ending[1:]
        return base[:-1] + ending


class Rahen(ConjugationType):
    """
    ラ行変格活用
    「ある」のみ。GodanZとほぼ同じ。
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return base == "ある" and (c_type_info == "ラ変" or "五段" in c_type_info)

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        return Godan.conjugate(base, ending)


class GodanI(Godan):
    """
    五段活用（イ音便）
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        is_i_onbin = (
            base[-1] in "くぐ" and not base.endswith(("行く", "逝く"))
        ) or base.endswith(Godan.POLITE)
        return Godan.conforms_to(pos_info, base, c_type_info) and is_i_onbin


class GodanZ(Godan):
    """
    五段活用（促音便）
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        is_z_onbin = (
            (base[-1] in "つるう" or base.endswith(("行く", "逝く")))
            and (not base.endswith(Godan.POLITE))
            and (not base.endswith(("問う", "請う")))
        )
        return Godan.conforms_to(pos_info, base, c_type_info) and is_z_onbin


class GodanN(Godan):
    """
    五段活用（撥音便）
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        is_n_onbin = base[-1] in "ぬぶむ"
        return Godan.conforms_to(pos_info, base, c_type_info) and is_n_onbin


class GodanU(Godan):
    """
    五段活用（ウ音便）

    Notes
    -----
    There are only a few, such as 問う and 請う.
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        is_u_onbin = base.endswith(("問う", "請う"))
        return Godan.conforms_to(pos_info, base, c_type_info) and is_u_onbin


class Ichidan(ConjugationType):
    """
    一段活用
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return "一段" in c_type_info

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        return base[:-1] + ending


class Kahen(ConjugationType):
    """
    カ行変格活用
    """

    TRANS = str.maketrans({"き": "来", "く": "来", "こ": "来"})

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return "カ" in c_type_info and "変" in c_type_info

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        if base.endswith("来る"):
            ending = ending.translate(Kahen.TRANS)
        return base[:-2] + ending


class Sahen(ConjugationType):
    """
    サ行変格活用
    """

    TRANS = str.maketrans(
        {
            "さ": "ざ",
            "し": "じ",
            "す": "ず",
            "せ": "ぜ",
        }
    )

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return "サ" in c_type_info and "変" in c_type_info

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        if base.endswith("ずる"):
            ending = ending.translate(Sahen.TRANS)
        return base[:-2] + ending


class Adjective(ConjugationType):
    """
    形容詞活用
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return pos.Adjective.conforms_to(pos_info)

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        return base[:-1] + ending


class AuxiliaryDa(ConjugationType):
    """
    助動詞「だ」
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return pos.Auxiliary.conforms_to(pos_info) and base == "だ"

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        return ending


class AuxiliaryNai(ConjugationType):
    """
    助動詞「ない」
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return pos.Auxiliary.conforms_to(pos_info) and base == "ない"

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        return Adjective.conjugate(base, ending)


class Nothing(ConjugationType):
    """
    No conjugation
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return c_type_info == ""

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        raise CannotConjugateError


class Unknown(ConjugationType):
    """
    Unknown conjugation form
    """

    @staticmethod
    def conforms_to(pos_info: List[str], base: str, c_type_info: str) -> bool:
        return False

    @staticmethod
    def conjugate(base: str, ending: str) -> str:
        raise CannotConjugateError


ALL_CTYPE: List[Type[ConjugationType]] = [
    Rahen,
    GodanI,
    GodanZ,
    GodanN,
    GodanU,
    Godan,  # get_normalized_ctype で先頭から順に判定するため五段活用の最後におく
    Ichidan,
    Kahen,
    Sahen,
    Adjective,
    AuxiliaryDa,
    AuxiliaryNai,
    Nothing,
]


def get_normalized_ctype(
    pos_info: List[str], base: str, c_type_info: str
) -> ConjugationType:
    for ctype in ALL_CTYPE:
        if ctype.conforms_to(pos_info, base, c_type_info):
            return ctype(value=c_type_info)
    return Unknown(value=c_type_info)



from abc import ABCMeta, abstractmethod
from typing import List, Type


class PartOfSpeech(metaclass=ABCMeta):
    """
    See Also
    --------
    ``https://www.sketchengine.eu/tagset-jp-mecab/``
    """

    def __init__(self, value: List[str]) -> None:
        self.value = value

    @staticmethod
    @abstractmethod
    def conforms_to(pos_info: List[str]) -> bool:
        pass

    def __str__(self) -> str:
        return self.__class__.__name__ + "(" + str(self.value) + ")"


class AdjectivalNoun(PartOfSpeech):
    """
    形状詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0] == "形状詞"


class Adjective(PartOfSpeech):
    """
    形容詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        is_suffix_type = (
            len(pos_info) > 1 and pos_info[0] == "接尾辞" and "形容詞" in pos_info[1]
        )
        return is_suffix_type or pos_info[0] == "形容詞"


class Adnominal(PartOfSpeech):
    """
    連体詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return any(pos.startswith("連体詞") for pos in pos_info)


class Adverb(PartOfSpeech):
    """
    副詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0] == "副詞"


class Auxiliary(PartOfSpeech):
    """
    助動詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0] in ("助動詞", "判定詞")


class Conjunction(PartOfSpeech):
    """
    接続詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0] == "接続詞"


class Interjection(PartOfSpeech):
    """
    感動詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0] in ("感動詞", "フィラー")


class Noun(PartOfSpeech):
    """
    名詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0].endswith("名詞")


class Particle(PartOfSpeech):
    """
    助詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0] == "助詞"


class EndingParticle(Particle):
    """
    終助詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        is_particle = Particle.conforms_to(pos_info)
        is_ending = len(pos_info) > 1 and "終助詞" in pos_info[1]
        return is_particle and is_ending


class Prefix(PartOfSpeech):
    """
    接頭辞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0].startswith("接頭")


class Suffix(PartOfSpeech):
    """
    接尾辞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return not Adjective.conforms_to(pos_info) and pos_info[0] == "接尾辞"


class Symbol(PartOfSpeech):
    """
    記号
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0] in ("記号", "補助記号", "空白", "特殊")


class Verb(PartOfSpeech):
    """
    動詞
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0] == "動詞"


class Other(PartOfSpeech):
    """
    その他
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return pos_info[0] in ("その他", "未定義語", "未知語")


class Unknown(PartOfSpeech):
    """
    Unknown part-of-speech
    """

    @staticmethod
    def conforms_to(pos_info: List[str]) -> bool:
        return False


ALL_POS: List[Type[PartOfSpeech]] = [
    AdjectivalNoun,
    Adjective,
    Adnominal,
    Adverb,
    Auxiliary,
    Conjunction,
    Interjection,
    Noun,
    EndingParticle,
    Particle,
    Prefix,
    Suffix,
    Symbol,
    Verb,
    Other,
]


def get_normalized_pos(pos_info: List[str]) -> PartOfSpeech:
    for pos in ALL_POS:
        if pos.conforms_to(pos_info):
            return pos(value=pos_info)
    return Unknown(value=pos_info)



import subprocess
from pathlib import Path
from typing import List

from jadoc.errors import MecabConfigError


def _mecab_config_dicdir(mecab_config: str) -> Path:
    """Get the result of executing ``mecab-config --dicdir``.

    Parameters
    ----------
    mecab_config : str
        Executable path of mecab-config.

    Returns
    -------
    Path
        The path where MeCab dictionary directories are stored.
    """
    try:
        result = subprocess.run(
            [mecab_config, "--dicdir"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        path = result.stdout.decode("utf-8").splitlines()[0]
        return Path(path)
    except FileNotFoundError:
        raise MecabConfigError("`mecab-config` is not found.")


def get_dicdirs(mecab_config: str = "mecab-config") -> List[Path]:
    """Get MeCab dictionary directories.

    Parameters
    ----------
    mecab_config : str
        Executable path of mecab-config, by default "mecab-config".

    Returns
    -------
    List[Path]
        MeCab dictionary directories.
    """
    dicdirs = []
    for path in _mecab_config_dicdir(mecab_config).glob("**/dicrc"):
        dicdirs.append(path.parent.resolve())
    return dicdirs



from pprint import pprint
from typing import Callable, Dict, List, Optional, Type

from .utils import debug_on
from .word.cform import (
    ConjugationForm,
    Gokan,
    IshiSuiryo,
    Katei,
    Meirei,
    Mizen,
    Rentai,
    Renyo,
    RenyoNi,
    RenyoOnbin,
    Shushi,
)
from .word.ctype import (
    ALL_CTYPE,
    Adjective,
    AuxiliaryDa,
    AuxiliaryNai,
    Godan,
    GodanI,
    GodanN,
    GodanU,
    GodanZ,
    Ichidan,
    Kahen,
    Rahen,
    Sahen,
)
from .word.word import Word


def _replace_with_vowel(hiragana_text: str) -> str:
    """
    Replace the first letter of ``hiragana_text`` with its vowel.

    Parameters
    ----------
    hiragana_text : str
        A string of Japanese hiragana characters.

    Returns
    -------
    str
        ``hiragana_text`` after replacement.

    Raises
    ------
    ValueError
        If the vowel of ``hiragana_text[0]`` is not found.

    Examples
    --------
    >>> assert _replace_with_vowel("こんにちは") == "oんにちは"
    >>> assert _replace_with_vowel("たのしい") == "aのしい"
    """
    for vowel, chars in Godan.JA_CHARS.items():
        if hiragana_text[0] in chars:
            return vowel + hiragana_text[1:]
    raise ValueError("Could not find the vowel of " + hiragana_text[0])


def show_details(func):
    def _show_details(self, word: Word, c_form: ConjugationForm) -> Word:
        before = str(word)
        new_word = func(self, word, c_form)
        after = str(new_word)
        if debug_on():
            all_args = str(word) + ", " + str(c_form)
            print(f"{self.__class__.__name__}.{func.__name__}({all_args}): ")
            if before != after:
                print(before)
                print(after)
        return new_word

    return _show_details


class Conjugation:
    def __init__(self, tokenize: Callable[[str], List[Word]]):
        self.tokenize = tokenize
        self._ending_dic = {ctype: {} for ctype in ALL_CTYPE}
        self._ending_dic[Godan] = self._generate_godan_ending_dic()
        self._ending_dic[Rahen] = self._generate_godan_ending_dic(renyo_onbin="っ")
        self._ending_dic[GodanI] = self._generate_godan_ending_dic(renyo_onbin="い")
        self._ending_dic[GodanZ] = self._generate_godan_ending_dic(renyo_onbin="っ")
        self._ending_dic[GodanN] = self._generate_godan_ending_dic(renyo_onbin="ん")
        self._ending_dic[GodanU] = self._generate_godan_ending_dic(renyo_onbin="う")
        self._ending_dic[Ichidan] = self._generate_ichidan_ending_dic()
        self._ending_dic[Kahen] = self._generate_kahen_ending_dic()
        self._ending_dic[Sahen] = self._generate_sahen_ending_dic()
        self._ending_dic[Adjective] = self._generate_adjective_ending_dic()
        self._ending_dic[AuxiliaryDa] = self._generate_auxiliary_da_ending_dic()
        self._ending_dic[AuxiliaryNai] = self._generate_adjective_ending_dic()
        if debug_on():
            pprint(self._ending_dic)

    def _generate_godan_ending_dic(
        self, renyo_onbin: Optional[str] = None
    ) -> Dict[Type[ConjugationForm], str]:
        words = self.tokenize("書かない。書こう。書きます。書く。書くとき。書けば。書け！")
        cforms = [Mizen, IshiSuiryo, Renyo, Shushi, Rentai, Katei, Meirei]
        endings = [
            _replace_with_vowel(word.surface[1:])
            for word in words
            if isinstance(word.c_type, Godan)
        ]
        assert len(cforms) == len(endings)
        ending_dic = {c: e for c, e in zip(cforms, endings)}
        if renyo_onbin is not None:
            ending_dic[RenyoOnbin] = renyo_onbin
        return ending_dic

    def _generate_ichidan_ending_dic(self) -> Dict[Type[ConjugationForm], str]:
        words = self.tokenize("起きない。起きよう。起きます。起きる。起きるとき。起きれば。起きろ！")
        cforms = [Mizen, IshiSuiryo, Renyo, Shushi, Rentai, Katei, Meirei]
        endings = [word.surface[2:] for word in words if type(word.c_type) == Ichidan]
        assert len(cforms) == len(endings)
        return {c: e for c, e in zip(cforms, endings)}

    def _generate_kahen_ending_dic(self) -> Dict[Type[ConjugationForm], str]:
        words = self.tokenize("持ってこない。持ってこよう。持ってきます。持ってくる。持ってくるとき。持ってくれば。持ってこい。")
        cforms = [Mizen, IshiSuiryo, Renyo, Shushi, Rentai, Katei, Meirei]
        endings = [word.surface for word in words if type(word.c_type) == Kahen]
        assert len(cforms) == len(endings)
        return {c: e for c, e in zip(cforms, endings)}

    def _generate_sahen_ending_dic(self) -> Dict[Type[ConjugationForm], str]:
        words = self.tokenize("すぐしない。読書しよう。します。する。するとき。すれば。読書せよ！")
        cforms = [Mizen, IshiSuiryo, Renyo, Shushi, Rentai, Katei, Meirei]
        endings = [word.surface for word in words if type(word.c_type) == Sahen]
        assert len(cforms) == len(endings)
        return {c: e for c, e in zip(cforms, endings)}

    def _generate_adjective_ending_dic(self) -> Dict[Type[ConjugationForm], str]:
        words = self.tokenize("美しかろう。美しく咲く。美しかった。美しい。美しいとき。美しければ。美しそう。")
        cforms = [IshiSuiryo, Renyo, RenyoOnbin, Shushi, Rentai, Katei, Gokan]
        endings = [word.surface[2:] for word in words if type(word.c_type) == Adjective]
        assert len(cforms) == len(endings)
        return {c: e for c, e in zip(cforms, endings)}

    def _generate_auxiliary_da_ending_dic(self) -> Dict[Type[ConjugationForm], str]:
        words = self.tokenize("鯖だろう。鯖である。鯖だった。鯖だ。鯖なの。鯖ならば。")
        cforms = [IshiSuiryo, Renyo, RenyoOnbin, Shushi, Rentai, Katei]
        endings = [word.surface for word in words if type(word.c_type) == AuxiliaryDa]
        assert len(cforms) == len(endings)
        ending_dic = {c: e for c, e in zip(cforms, endings)}
        ending_dic[RenyoNi] = "に"
        return ending_dic

    @show_details
    def conjugate(self, word: Word, c_form: ConjugationForm) -> Word:
        if not word.has_conjugation:
            return word

        ctype = type(word.c_type)
        cform = type(c_form)

        try:
            if cform == RenyoOnbin and RenyoOnbin not in self._ending_dic[ctype]:
                c_form = Renyo(value="連用形")
                cform = type(c_form)
            ending = self._ending_dic[ctype][cform]
            if not word.base.endswith("しい") and cform == Gokan:
                ending = "さ"  # 語幹-サ
        except KeyError:
            return word

        word.surface = ctype.conjugate(word.base, ending)
        word.c_form = c_form

        return word



class JadocError(Exception):
    """
    Base class for exceptions in this module.
    """

    pass


class CannotConjugateError(JadocError):
    """
    Raised when a word with no conjugation is conjugated.
    """

    pass


class MecabConfigError(JadocError):
    """
    Raised when the ``mecab-config`` command is not found.
    """

    pass


class NotFoundNodeFormatError(JadocError):
    """
    Raised when the MeCab ``node-format`` could not be set automatically.
    """

    pass


class InvalidTokenizerError(JadocError):
    """
    Raised when the tokenizer is not working properly.
    """

    pass



from typing import Dict, List, Optional, Union

from .conj import Conjugation
from .mecab.tokenizer import generate_tokenizer
from .utils import debug_on
from .word.cform import ConjugationForm, Mizen, Renyo, RenyoOnbin
from .word.ctype import Sahen
from .word.word import Word


def show_details(func):
    def _show_details(self, *args, **kwargs) -> None:
        before = self.simple_view()
        result = func(self, *args, **kwargs)
        after = self.simple_view()
        if debug_on():
            all_args = ", ".join(
                [str(a) for a in args] + [f"{k}={str(v)}" for k, v in kwargs.items()]
            )
            print(f"{self.__class__.__name__}.{func.__name__}({all_args}): ")
            if before != after:
                print(before)
                print(after)
        return result

    return _show_details


class Doc:
    def __init__(self, text: str, conjugation: Conjugation = None) -> None:
        if conjugation is None:
            conjugation = Conjugation(tokenize=generate_tokenizer())
        self.conjugation = conjugation
        self.words: List[Word] = self.conjugation.tokenize(text)
        if debug_on():
            print("Doc.__init__(): \n" + self.simple_view())

    def is_within_range(self, interval: Union[int, range]) -> bool:
        index_max = len(self.words) - 1
        if type(interval) == int:
            return 0 <= interval <= index_max
        else:
            return 0 <= interval.start <= (interval.stop - 1) <= index_max

    def get_text(self, interval: Optional[Union[int, range]] = None) -> str:
        if interval is None:
            interval = range(0, len(self.words))
        if type(interval) == int:
            interval = range(interval, interval + 1)

        surface = ""
        for i in interval:
            surface += self.words[i].surface
        return surface

    @show_details
    def retokenize(self, text: Optional[str] = None) -> None:
        if text is None:
            text = self.get_text()
        self.words = self.conjugation.tokenize(text)

    @show_details
    def _conjugate_irregularly(self, i: int, c_form: ConjugationForm) -> bool:
        """
        Returns
        -------
        True
            If this is an irregular case and conjugation has done.
        False
            Otherwise.
        """
        ctype = type(self.words[i].c_type)
        cform = type(c_form)

        # irregular case of 「ある」
        is_aru_mizen_case = all(
            [
                self.words[i].base == "ある",
                cform == Mizen,
                self.is_within_range(i + 1) and self.words[i + 1].base == "ない",
            ]
        )
        if is_aru_mizen_case:
            del self.words[i]
            return True

        # irregular case of 「する」
        is_suru_mizen_case = (
            ctype == Sahen and cform == Mizen and self.is_within_range(i + 1)
        )
        if is_suru_mizen_case:
            if self.words[i + 1].base in ("せる", "れる"):
                self.words[i].surface = "さ"
                self.words[i].c_form = Mizen(value="未然形")
                return True
            elif self.words[i + 1].base == "ぬ":
                self.words[i].surface = "せ"
                self.words[i].c_form = Mizen(value="未然形")
                return True

        return False

    @show_details
    def _conjugate_renyo_onbin(self, i: int) -> None:
        surface = self.words[i + 1].surface
        s = surface[0]

        renyo_onbin = RenyoOnbin(value="連用形-音便")
        self.conjugation.conjugate(self.words[i], renyo_onbin)

        if self.words[i].base[-1] in "ぬぶむ":
            s = s.replace("た", "だ").replace("て", "で")
        else:
            s = s.replace("だ", "た").replace("で", "て")
        self.words[i + 1].surface = s + surface[1:]

    @show_details
    def conjugate(self, i: int, c_form: ConjugationForm) -> None:
        if not self.is_within_range(i):
            return

        if self._conjugate_irregularly(i, c_form):
            return

        cform = type(c_form)

        next_is_td = (
            self.is_within_range(i + 1) and self.words[i + 1].surface[0] in "ただてで"
        )
        is_renyo_onbin_case = (cform == Renyo or cform == RenyoOnbin) and next_is_td
        if is_renyo_onbin_case:
            self._conjugate_renyo_onbin(i)
            return

        # normal case
        if cform == RenyoOnbin and not next_is_td:
            c_form = Renyo("連用形")
        self.conjugation.conjugate(self.words[i], c_form)

    @show_details
    def insert(self, i: int, words: Union[Word, List[Word]]) -> None:
        if type(words) == Word:
            words = [words]
        self.words[i:i] = words

        c_form = self.words[i - 1].c_form
        self.conjugate(i - 1, c_form)
        self.conjugate(i - 1 + len(words), c_form)

    @show_details
    def delete(self, interval: Union[int, range]) -> None:
        if type(interval) == int:
            interval = range(interval, interval + 1)

        if not self.is_within_range(interval):
            return

        c_form = self.words[interval.stop - 1].c_form
        del self.words[interval.start : interval.stop]
        self.conjugate(interval.start - 1, c_form)

    @show_details
    def update(
        self, interval: Union[int, range], words: Union[Word, List[Word]]
    ) -> None:
        if type(interval) == int:
            interval = range(interval, interval + 1)
        if not self.is_within_range(interval):
            return
        if type(words) == Word:
            words = [words]

        n = len(words)
        self.insert(interval.start, words)
        self.delete(range(interval.start + n, interval.stop + n))

    @show_details
    def update_surfaces(
        self, interval: Union[int, range], surfaces: Union[str, List[str]]
    ) -> None:
        if type(interval) == int:
            interval = range(interval, interval + 1)
        if not self.is_within_range(interval):
            return
        if type(surfaces) == str:
            surfaces = [surfaces]

        pre = self.get_text(range(0, interval.start))
        surface = "".join(surfaces)
        post = self.get_text(range(interval.stop, len(self.words)))
        text = pre + surface + post
        self.retokenize(text)

    def simple_view(self, sep: str = "/") -> str:
        tokens = []
        for word in self.words:
            token = word.surface + "[" + str(word.pos)[:3] + "]"
            if word.has_conjugation:
                token += "(" + str(word.c_type)[:3] + ";" + str(word.c_form)[:3] + ")"
            tokens.append(token)

        return sep.join(tokens)

    def to_word_list(self) -> List[Dict[str, str]]:
        return [word.to_dict() for word in self.words]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jadoc', 'jadoc.mecab', 'jadoc.word']

package_data = \
{'': ['*']}

install_requires = \
['mecab-python3>=1.0.3,<2.0.0']

setup_kwargs = {
    'name': 'jadoc',
    'version': '0.2.4',
    'description': 'Tokenizes Japanese documents to enable CRUD operations.',
    'long_description': '# Jadoc: Tokenizes Japanese Documents to Enable CRUD Operations\n\n[![PyPI Version](https://img.shields.io/pypi/v/jadoc.svg)](https://pypi.org/pypi/jadoc/)\n[![Python Versions](https://img.shields.io/pypi/pyversions/jadoc.svg)](https://pypi.org/pypi/jadoc/)\n[![License](https://img.shields.io/pypi/l/jadoc.svg)](https://github.com/poyo46/jadoc/blob/main/LICENSE)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n\n## Installation\n\n**Install MeCab**\n\nMeCab is required for Jadoc to work.\nIf it is not already installed, [install MeCab](https://taku910.github.io/mecab/) first.\n\n**Install Jadoc**\n\n```console\n$ pip install jadoc\n```\n\n## Examples\n\n```python\nfrom jadoc.doc import Doc\n\n\ndoc = Doc("本を書きました。")\n\n# print surface forms of the tokens.\nsurfaces = [word.surface for word in doc.words]\nprint("/".join(surfaces))  # 本/を/書き/まし/た/。\n\n# print plain text\nprint(doc.get_text())  # 本を書きました。\n\n# delete a word\ndoc.delete(3)  # Word conjugation will be done as needed.\nprint(doc.get_text())  # 本を書いた。\n\n# update a word\nword = doc.conjugation.tokenize("読む")\n# In addition to conjugation, transform the peripheral words as needed.\ndoc.update(2, word)\nprint(doc.get_text())  # 本を読んだ。\n```\n',
    'author': 'poyo46',
    'author_email': 'poyo4rock@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/poyo46/jadoc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from abc import ABCMeta, abstractmethod
from typing import List, Type

from .pos import Adjective, Auxiliary


class ConjugationForm(metaclass=ABCMeta):
    """
    See Also
    --------
    ``https://www.sketchengine.eu/tagset-jp-mecab/``
    """

    def __init__(self, value: str) -> None:
        self.value = value

    @staticmethod
    @abstractmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        pass

    def __str__(self) -> str:
        return self.__class__.__name__ + "(" + self.value + ")"


class Mizen(ConjugationForm):
    """
    未然形
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        return (
            not IshiSuiryo.conforms_to(surface, pos_info, c_form_info)
            and "未然" in c_form_info
        )


class IshiSuiryo(ConjugationForm):
    """
    意志推量形
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        is_daro = surface == "だろ" and c_form_info == "未然形"
        is_desyo = surface == "でしょ" and c_form_info == "未然形"
        is_irregular = is_daro or is_desyo
        return "意志推量" in c_form_info or "未然ウ接続" in c_form_info or is_irregular


class Renyo(ConjugationForm):
    """
    連用形
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        return "連用" in c_form_info


class RenyoOnbin(ConjugationForm):
    """
    連用形-音便
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        is_onbin = (
            "音便" in c_form_info or "連用タ接続" in c_form_info or "連用ゴザイ接続" in c_form_info
        )
        return Renyo.conforms_to(surface, pos_info, c_form_info) and is_onbin


class RenyoNi(ConjugationForm):
    """
    連用形-ニ
    only ``ctype.AuxiliaryDa`` has this conjugation form.
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        is_ni = "ニ" in c_form_info
        return Renyo.conforms_to(surface, pos_info, c_form_info) and is_ni


class Shushi(ConjugationForm):
    """
    終止形
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        return "終止" in c_form_info or "基本" in c_form_info


class Rentai(ConjugationForm):
    """
    連体形
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        return "連体" in c_form_info or "体言接続" in c_form_info


class Katei(ConjugationForm):
    """
    仮定形
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        return "仮定" in c_form_info


class Meirei(ConjugationForm):
    """
    命令形
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        return "命令" in c_form_info


class Gokan(ConjugationForm):
    """
    語幹
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        pos_condition = Adjective.conforms_to(pos_info) or Auxiliary.conforms_to(
            pos_info
        )
        return pos_condition and c_form_info in ("ガル接続", "語幹-一般")


class Nothing(ConjugationForm):
    """
    No conjugation
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        return c_form_info == ""


class Unknown(ConjugationForm):
    """
    Unknown conjugation form
    """

    @staticmethod
    def conforms_to(surface: str, pos_info: List[str], c_form_info: str) -> bool:
        return False


ALL_CFORM: List[Type[ConjugationForm]] = [
    IshiSuiryo,
    Mizen,
    RenyoNi,
    RenyoOnbin,
    Renyo,
    Shushi,
    Rentai,
    Katei,
    Meirei,
    Gokan,
    Nothing,
]


def get_normalized_cform(
    surface: str, pos_info: List[str], c_form_info: str
) -> ConjugationForm:
    for cform in ALL_CFORM:
        if cform.conforms_to(surface, pos_info, c_form_info):
            return cform(value=c_form_info)
    return Unknown(value=c_form_info)



__title__ = "jadoc"
__description__ = "Tokenizes Japanese documents to enable CRUD operations."
__url__ = "https://github.com/poyo46/jadoc"
__version__ = "0.2.4"
__author__ = "poyo46"
__author_email__ = "poyo4rock@gmail.com"
__license__ = "Apache-2.0"
__copyright__ = "Copyright 2021 poyo46"
